%
% Programmed and algorihtm by Chanwoo Kim (chanwook@cs.cmu.edu)
%
% This package is used for the INTERSPEECH 2009 paper
%
% Imporrant: The ITD threshold is for "4 cm" distance microphone and 16 kHz sampling
% rate. If the microphone distance and the sampling rate is different from
% this configuration you need to change dTH in the code.
%
% You can do select between PD and PDCW  by changning bPDCW option
%
% if bPDCW == 1 it works as PDCW if not it works as PD
%
%

function [] = AutomaticITD(szOutFileName, szInFileName, bAuto)
    %
    % 75 ms (0.1 s) frame length
    dFrameLen = 0.075;
    %
    % ITD threshold is for "4 cm" microphone with 16 kHz sampling rate case
    %
    dMicDist = 0.04;
    
    %
    % Search range setting for automatic ITD system.
    %
    % Search is from 3 degree up to 45 degree
    %
    adITD_TH_SEARCH = (dMicDist * sin(3 * pi / 180) / 340 * 16000) : 0.05 : (dMicDist * sin(45 * pi / 180) / 340 * 16000);
    
    %
    % In the fixed ITD system, it is assumed that the interference is 45
    % degree and 22.5 degree is selected as the threshold
    %
    dFixedITD       = (dMicDist * sin(22.5 * pi / 180) / 340 * 16000);
    
    %
    % This part for mswav
    %
    ad_x = wavread(szInFileName)';
    
    dSampRate      = 16000;
    iSpeechLen     = length(ad_x);
    iFrameLen      = round(dSampRate * dFrameLen);
    iFramePeriod   = round(dSampRate * dFrameLen / 2); % To meet the OLA constraint
    iFFTSize       = 2^ceil(log(iFrameLen) / log(2));
    dEta         = 1e-2;
	dPowerFactor = 1;
    
    dPowerCoeff = 1 / 15;

    dOrgStd  = std(ad_x(1, :));
    ad_X_L   = zeros(1, iFFTSize);
    ad_X_R   = zeros(1, iFFTSize);
    ad_y     = zeros(1, length(ad_x));

    iNumFrames  = floor((iSpeechLen - iFrameLen) / iFramePeriod) + 1;
    aad_X_tilde = zeros(iFFTSize, iNumFrames);
    aad_X_tildeComp = zeros(iFFTSize, iNumFrames);
    aad_mu      = zeros(iFFTSize / 2, iNumFrames);
    aad_mu_g    = zeros(iFFTSize / 2, iNumFrames);
    
    aad_mu_T    = zeros(iFFTSize / 2, iNumFrames);
    
    ad_R_T   = zeros(iNumFrames, length(adITD_TH_SEARCH));
    ad_R_I   = zeros(iNumFrames, length(adITD_TH_SEARCH));
    
    aad_X_Bar = zeros(iFFTSize, iNumFrames);
    aad_ITD   = zeros(iFFTSize / 2, iNumFrames);
  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Analysis
    %
    iFI = 0;    
    adBuffer  = zeros(iFFTSize, 1);
    iNumFilts = 40;
    aad_H     = ComputeFilterResponse(iNumFilts, iFFTSize);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Frame-by-frame processing
    %
    for i = 0 : iFramePeriod : iSpeechLen - 1 -  iFrameLen,
        iFI = iFI + 1;

        ad_x_L_Frame = ad_x(1, i + 1 : i + iFrameLen) .* hamming(iFrameLen)';
        ad_x_R_Frame = ad_x(2, i + 1 : i + iFrameLen) .* hamming(iFrameLen)';
        ad_X_L       = fft(ad_x_L_Frame, iFFTSize);
        ad_X_R       = fft(ad_x_R_Frame, iFFTSize);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        % Now, let's calculate the phase difference
        %
        ad_X_LAngle  = (angle(ad_X_L)); 
        ad_X_RAngle  = (angle(ad_X_R ));
        adPhaseDiff1 = abs(ad_X_LAngle - ad_X_RAngle);
        adPhaseDiff2 = abs(ad_X_LAngle - ad_X_RAngle - 2 * pi);
        adPhaseDiff3 = abs(ad_X_LAngle - ad_X_RAngle + 2 * pi);
        adPhaseDiff  = min([adPhaseDiff1; adPhaseDiff2; adPhaseDiff3]);

        ad_X_Bar = (ad_X_L + ad_X_R) / 2;

        aad_X_Bar(:, iFI) = ad_X_Bar.';

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        % ITD calculation from the phase difference
        %
        ad_ITD(1) = 0;
        ad_ITD(2 : iFFTSize / 2) = (adPhaseDiff(2 : iFFTSize / 2) ./ (2 * pi * (1 : iFFTSize / 2 - 1) / iFFTSize))';

        aad_ITD(:, iFI) = ad_ITD';

        ad_ITDBuffer = ad_ITD;

        iSearchIndex = 0;

        for dITD_TH = adITD_TH_SEARCH

            iSearchIndex = iSearchIndex + 1;

            ad_mu_T          = zeros(iFFTSize / 2, 1);
            ad_mu_I          = zeros(iFFTSize / 2, 1);

            aiIndex          = find(ad_ITD < dITD_TH );

            ad_mu_T(aiIndex) = 1; 
            aiIndex          = find(ad_ITD > dITD_TH ); 

            ad_mu_I(aiIndex) = 1;
            ad_mu_Ttemp = ad_mu_T;

            ad_mu_T = ad_mu_Ttemp;

            ad_mu_T = max(ad_mu_T, dEta);
            ad_mu_I = max(ad_mu_I, dEta);

            ad_X_T = ad_X_Bar(1 : iFFTSize / 2) .*  ad_mu_T';
            ad_X_I = ad_X_Bar(1 : iFFTSize / 2) .*  ad_mu_I';

            ad_R_T(iFI, iSearchIndex) = sum(abs(ad_X_T .^ 2)) ^ (dPowerCoeff);
            ad_R_I(iFI, iSearchIndex) = sum(abs(ad_X_I .^ 2)) ^ (dPowerCoeff);

        end
            ad_X_PD = ad_X_Bar(1 : iFFTSize / 2) .*  aad_mu( :, iFI)';
            adPDPower(iFI) = sum((abs(ad_X_PD)) .^ 2);
    end

    iSearchIndex = 0;
    for dITD_TH = adITD_TH_SEARCH
        iSearchIndex = iSearchIndex + 1;

        adRho(iSearchIndex) = (mean(ad_R_T(:, iSearchIndex) .* ad_R_I(:, iSearchIndex)) - mean(ad_R_T(:, iSearchIndex)) * mean(ad_R_I(:, iSearchIndex))) / std(ad_R_T(:, iSearchIndex)) / std(ad_R_I(:, iSearchIndex));
    end

    [dVal iITDIndex] = min(abs(adRho));
        
        
   if (bAuto == 1)

        dOptITD_TH = adITD_TH_SEARCH(iITDIndex);
    
        acBuffer = sprintf('Obtained optimal ITD threshold is %f samples', dOptITD_TH);
        disp(acBuffer)
        
    else % if bAuto = 0
        
        acBuffer = sprintf('Fixed ITD threshold is %f samples', dFixedITD);
        disp(acBuffer)
        dOptITD_TH = dFixedITD;
    end
    
    iFI = 0;
    for i = 0 : iFramePeriod : iSpeechLen - 1 -  iFrameLen, 
        iFI = iFI + 1;
        
        ad_ITD   = aad_ITD(:, iFI)';
        ad_X_Bar = aad_X_Bar(:, iFI).';
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        %  time-frequency mask /mu
        %
        ad_mu            = zeros(iFFTSize / 2, 1);
        aiIndex          = find(ad_ITD < dOptITD_TH);
        ad_mu(aiIndex)   = 1;
        
        ad_mu            = max(ad_mu, dEta);
        aad_mu( :, iFI)  = ad_mu;
        
        iSearchIndex = 0;
    
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        % The processing using PD (Phase Difference)
        %
        ad_mu_sym = [aad_mu(:, iFI); flipud(aad_mu( :, iFI))];

        ad_mu_g_sym_comp = 1 - ad_mu_sym;
        ad_mu_g_sym_comp = max(ad_mu_g_sym_comp, dEta);

        ad_mu     = max(ad_mu, dEta);

        aad_X_tilde( :, iFI)     = (ad_X_Bar .*  ad_mu_sym')';   
        aad_X_tildeComp( :, iFI) = (ad_X_Bar .*  ad_mu_g_sym_comp')';
    end
        
    ad_y     = zeros(1, length(ad_x));
    ad_yComp = zeros(1, length(ad_x));
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Resynthesis using OLA
    %
    iFI = 0;
    for i = 0 : iFramePeriod : iSpeechLen - 1 -  iFrameLen,
         iFI = iFI + 1;
         aad_X_tilde( :, iFI) = aad_X_tilde( :, iFI); 
         adBuffer = fliplr(ifft(aad_X_tilde( :, iFI))');
         adBuffer2 = fliplr(ifft(aad_X_tildeComp( :, iFI))');
         ad_y(i + 1 : i + iFrameLen)     = ad_y(i + 1 : i + iFrameLen) + adBuffer(1 : iFrameLen);
         ad_yComp(i + 1 : i + iFrameLen) = ad_yComp(i + 1 : i + iFrameLen) + adBuffer2(1 : iFrameLen);
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Selecting only the real part
    %
    ad_y     = real(ad_y);
    ad_yComp = real(ad_yComp);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Removing the first and the last frames.. there were not overlapped
    %
    ad_y(length(ad_y) - iFrameLen : length(ad_y)) = [];
    ad_y(1 : iFrameLen)                           = [];
    
    ad_yComp(length(ad_yComp) - iFrameLen : length(ad_yComp)) = [];
    ad_yComp(1 : iFrameLen)                                   = [];
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % If you want to see the mask please use the following imagescs
    %
    if 0
        figure
        imagesc(aad_mu); axis xy;
    end   
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Writing the data
    %
    %
    % Writing the data in MS wave format
    %
    wavwrite(ad_y / max(abs(ad_y(:))) / 1.1, 16000, 16, szOutFileName)
   
end

